import pandas as pd
import numpy as np


class ValidationError(Exception):
    """Raised when an error related to the validation is encountered.
    """


def validate_ref_point_with_ideal_and_nadir(
    dimensions_data: pd.DataFrame, reference_point: pd.DataFrame
):
    validate_ref_point_dimensions(dimensions_data, reference_point)
    validate_ref_point_data_type(reference_point)
    validate_ref_point_with_ideal(dimensions_data, reference_point)
    validate_with_ref_point_nadir(dimensions_data, reference_point)


def validate_ref_point_with_ideal(
    dimensions_data: pd.DataFrame, reference_point: pd.DataFrame
):
    validate_ref_point_dimensions(dimensions_data, reference_point)
    ideal_fitness = dimensions_data.loc["ideal"] * dimensions_data.loc["minimize"]
    ref_point_fitness = reference_point * dimensions_data.loc["minimize"]
    if not (ideal_fitness <= ref_point_fitness).all(axis=None):
        problematic_columns = ideal_fitness.index[
            (ideal_fitness > ref_point_fitness).values.tolist()[0]
        ].values
        msg = (
            f"Reference point should be worse than or equal to the ideal point\n"
            f"The following columns have problematic values: {problematic_columns}"
        )
        raise ValidationError(msg)


def validate_with_ref_point_nadir(
    dimensions_data: pd.DataFrame, reference_point: pd.DataFrame
):
    validate_ref_point_dimensions(dimensions_data, reference_point)
    nadir_fitness = dimensions_data.loc["nadir"] * dimensions_data.loc["minimize"]
    ref_point_fitness = reference_point * dimensions_data.loc["minimize"]
    if not (ref_point_fitness <= nadir_fitness).all(axis=None):
        problematic_columns = nadir_fitness.index[
            (nadir_fitness < ref_point_fitness).values.tolist()[0]
        ].values
        msg = (
            f"Reference point should be better than or equal to the nadir point\n"
            f"The following columns have problematic values: {problematic_columns}"
        )
        raise ValidationError(msg)


def validate_ref_point_dimensions(
    dimensions_data: pd.DataFrame, reference_point: pd.DataFrame
):
    if not dimensions_data.shape[1] == reference_point.shape[1]:
        msg = (
            f"There is a mismatch in the number of columns of the dataframes.\n"
            f"Columns in dimensions data: {dimensions_data.columns}\n"
            f"Columns in the reference point provided: {reference_point.columns}"
        )
        raise ValidationError(msg)
    if not all(dimensions_data.columns == reference_point.columns):
        msg = (
            f"There is a mismatch in the column names of the dataframes.\n"
            f"Columns in dimensions data: {dimensions_data.columns}\n"
            f"Columns in the reference point provided: {reference_point.columns}"
        )
        raise ValidationError(msg)


def validate_ref_point_data_type(reference_point: pd.DataFrame):
    for dtype in reference_point.dtypes:
        if not pd.api.types.is_numeric_dtype(dtype):
            msg = (
                f"Type of data in reference point dataframe should be numeric.\n"
                f"Provided datatype: {dtype}"
            )
            raise ValidationError(msg)


def validate_specified_solutions(indices: np.ndarray, n_solutions: int) -> None:
    """
    Validate the Decision maker's choice of preferred/non-preferred solutions.

    Args:
        indices (np.ndarray): Index/indices of preferred solutions specified by the Decision maker.
        n_solutions (int): Number of solutions in total.

    Returns:

    Raises:
        ValidationError: In case the preference is invalid.

    """

    if indices.shape[0] < 1:
        raise ValidationError("Please specify at least one (non-)preferred solution.")
    if not isinstance(indices, (np.ndarray, list)):
        raise ValidationError("Please specify index/indices of (non-)preferred solutions in a list, even if there is only "
                           "one.")
    if not all(0 <= i <= (n_solutions - 1) for i in indices):
        msg = "indices of (non-)preferred solutions should be between 0 and {}. Current indices are {}." \
            .format(n_solutions - 1, indices)
        raise ValidationError(msg)


def validate_bounds(dimensions_data: pd.DataFrame, bounds: np.ndarray, n_objectives: int) -> None:
    """
    Validate the Decision maker's desired lower and upper bounds for objective values.

    Args:
        dimensions_data (pd.DataFrame): DataFrame including information whether an objective is minimized or
        maximized, for each objective. In addition, includes ideal and nadir vectors.
        bounds (np.ndarray): Desired lower and upper bounds for each objective.
        n_objectives (int): Number of objectives in problem.

    Returns:

    Raises:
        ValidationError: In case desired bounds are invalid.

    """

    if not isinstance(bounds, np.ndarray):
        msg = "Please specify bounds as a numpy array. Current type: {}.".format(type(bounds))
        raise ValidationError(msg)
    if len(bounds) != n_objectives:
        msg = "Length of 'bounds' ({}) must be the same as number of objectives ({}).".format(len(bounds), n_objectives)
        raise ValidationError(msg)
    if not all(isinstance(b, (np.ndarray, list)) for b in bounds):
        print(type(bounds[0]))
        msg = "Please give bounds for each objective in a list."
        raise ValidationError(msg)
    if any(len(b) != 2 for b in bounds):
        msg = "Length of each item of 'bounds' must 2, containing the lower and upper bound for an objective."
        raise ValidationError(msg)
    if any(b[0] > b[1] for b in bounds):
        msg = "Lower bound cannot be greater than upper bound. Please specify lower bound first, then upper bound."
        raise ValidationError(msg)

    # check that bounds are within ideal and nadir points for each objective
    for i, b in enumerate(bounds):
        if dimensions_data.loc['minimize'].values.tolist()[i] == 1:  # minimized objectives
            if dimensions_data.loc['ideal'].values.tolist()[i] is not None:
                if b[0] < dimensions_data.loc['ideal'].values.tolist()[i]:
                    msg = "Lower bound cannot be lower than ideal value for objective. Ideal vector: {}." \
                        .format(dimensions_data.loc['ideal'].values.tolist())
                    raise ValidationError(msg)
            if dimensions_data.loc['nadir'].values.tolist()[i] is not None:
                if b[1] > dimensions_data.loc['nadir'].values.tolist()[i]:
                    msg = "Upper bound cannot be higher than nadir value for objective. Nadir vector: {}." \
                        .format(dimensions_data.loc['nadir'].values.tolist())
                    raise ValidationError(msg)

        else:  # maximized objectives:
            if dimensions_data.loc['ideal'].values.tolist()[i] is not None:
                if b[1] > dimensions_data.loc['ideal'].values.tolist()[i]:
                    msg = "Upper bound cannot be higher than ideal value for objective. Ideal vector: {}." \
                        .format(dimensions_data.loc['ideal'].values.tolist())
                    raise ValidationError(msg)
            if dimensions_data.loc['nadir'].values.tolist()[i] is not None:
                if b[0] < dimensions_data.loc['nadir'].values.tolist()[i]:
                    msg = "Lower bound cannot be lower than nadir value for objective. Nadir vector: {}." \
                        .format(dimensions_data.loc['nadir'].values.tolist())
                    raise ValidationError(msg)




"""Provides implementations that convert one type of preference information to another."""

import numpy as np


def classification_to_reference_point(
    classification_preference: dict, ideal: np.ndarray, nadir: np.ndarray
) -> dict:
    """Convert classification type of preference (e.g. NIMBUS) to reference point
        preference.

    Args:
        classification_preference (dict): A dict containing keys 'current solution',
            'levels', and 'classifications'. Read the NIMBUS paper for more details.
        ideal (np.ndarray): The ideal point of the problem.
        nadir (np.ndarray): The nadir point of the problem.

    Returns:
        dict: The preference in the form of a reference point. Contains one key:
            "reference point", which maps to the preference in a numpy array structure.
    """
    z_bar = np.zeros_like(nadir, dtype=float)

    improve_inds = np.where(
        np.array(classification_preference["classifications"]) == "<"
    )[0]
    acceptable_inds = np.where(
        np.array(classification_preference["classifications"]) == "="
    )[0]
    free_inds = np.where(np.array(classification_preference["classifications"]) == "0")[
        0
    ]
    improve_until_inds = np.where(
        np.array(classification_preference["classifications"]) == "<="
    )[0]
    impaire_until_inds = np.where(
        np.array(classification_preference["classifications"]) == ">="
    )[0]

    z_bar[improve_inds] = ideal[improve_inds]
    z_bar[improve_until_inds] = classification_preference["levels"][improve_until_inds]
    z_bar[acceptable_inds] = classification_preference["current solution"][
        acceptable_inds
    ]
    z_bar[impaire_until_inds] = classification_preference["levels"][impaire_until_inds]
    z_bar[free_inds] = nadir[free_inds]

    return {"reference point": z_bar}



"""A file to contain different kinds of lattice generation algorithms.

"""

import numba
import numpy as np


@numba.njit()
def fibonacci_sphere(samples: int = 1000) -> np.ndarray:
    """Generate a very even lattice of points on a 3d sphere using the fibonacci sphere
        or fibonacci spiral algorithm.

    Args:
        samples (int, optional): Number of points to be generated. Defaults to 1000.

    Returns:
        np.ndarray: The lattice of points as a 2-D (samples, 3) numpy array.
    """
    points = np.zeros((samples, 3), dtype=np.float_)
    phi = np.pi * (3 - np.sqrt(5))  # golden angle in radians

    for i in range(samples):
        points[i, 1] = 1 - (i / (samples - 1)) * 2  # y goes from 1 to -1
        radius = np.sqrt(1 - points[i, 1] ** 2)  # radius at y

        theta = phi * i  # golden angle increment

        points[i, 0] = np.cos(theta) * radius
        points[i, 2] = np.sin(theta) * radius
    return points



"""This module implements methods for solving scalar valued functions.

"""

__all__ = ["DiscreteMinimizer", "ScalarMethod", "ScalarMinimizer", "ScalarSolverException"]

from desdeo_tools.solver.ScalarSolver import DiscreteMinimizer, ScalarMethod, ScalarMinimizer, ScalarSolverException



from abc import abstractmethod
from typing import Union

import numpy as np


class GLIDEError(Exception):
    """Raised when an error related to the ASF classes is encountered.

    """


class GLIDEBase:
    """
    Implements the non-differentiable variant of GLIDE-II as proposed in
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Note:
        Additional contraints produced by the GLIDE-II formulation are implemented
        such that if the returned values are negative, the corresponding constraint is
        violated. The returned value may be positive. In such cases, the returned value
        is a measure of how close or far the corresponding feasible solution is from
        violating the constraint.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):

        self.has_additional_constraints = False
        self.utopian = utopian
        self.nadir = nadir
        self.rho = rho
        self.required_keys: dict = {}

    def __call__(self, objective_vector: np.ndarray, preference: dict) -> np.ndarray:
        """Evaluate the scalarization function value based on objective vectors and
        DM preference.

        Args:
            objective_vector (np.ndarray): 2-dimensional array of objective values of solutions.
            preference (dict): The preference given by the decision maker. The required
                dictionary keys and their meanings can be found in self.required_keys variable.

        Returns:
            np.ndarray: The scalarized value obtained by using GLIDE-II over
                objective_vector.
        """
        self.preference = preference
        self.objective_vector = objective_vector

        f_minus_q = np.atleast_2d(objective_vector - self.q)
        mu = np.atleast_2d(self.mu)
        I_alpha = self.I_alpha

        max_term = np.max(mu[:, I_alpha] * f_minus_q[:, I_alpha], axis=1)
        sum_term = self.rho * np.sum(self.w * f_minus_q, axis=1)
        return max_term + sum_term

    def evaluate_constraints(
        self, objective_vector: np.ndarray, preference: dict
    ) -> Union[None, np.ndarray]:
        """Evaluate the additional contraints generated by the GLIDE-II formulation.

        Note:
            Additional contraints produced by the GLIDE-II formulation are implemented
            such that if the returned values are negative, the corresponding constraint is
            violated. The returned value may be positive. In such cases, the returned value
            is a measure of how close or far the corresponding feasible solution is from
            violating the constraint.

        Args:
            objective_vector (np.ndarray): [description]
            preference (dict): [description]

        Returns:
            Union[None, np.ndarray]: [description]
        """
        if not self.has_additional_constraints:
            return None

        self.preference = preference
        self.objective_vector = objective_vector

        constraints = (
            self.epsilon[self.I_epsilon]
            + self.s_epsilon * self.delta_epsilon[self.I_epsilon]
            - objective_vector[:, self.I_epsilon]
        )

        return constraints

    @property
    @abstractmethod
    def I_alpha(self):
        pass

    @property
    @abstractmethod
    def I_epsilon(self):
        pass

    @property
    @abstractmethod
    def mu(self):
        pass

    @property
    @abstractmethod
    def q(self):
        pass

    @property
    @abstractmethod
    def w(self):
        pass

    @property
    @abstractmethod
    def epsilon(self):
        pass

    @property
    @abstractmethod
    def s_epsilon(self):
        pass

    @property
    @abstractmethod
    def delta_epsilon(self):
        pass


class reference_point_method_GLIDE(GLIDEBase):
    """
    Implements the reference point method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):
        super().__init__(utopian=utopian, nadir=nadir, rho=rho)
        self.has_additional_constraints = False
        self.__I_alpha = np.full_like(
            utopian, dtype=np.bool_, fill_value=True
        ).flatten()
        self.__I_epsilon = np.full_like(
            utopian, dtype=np.bool_, fill_value=False
        ).flatten()
        self.__w = 1
        self.__mu = 1 / (nadir - utopian)
        self.required_keys = {
            "reference point": (
                "Used to calculate the direction of improvement: "
                "a line parallel to the nadir-utopian vector "
                "and passing through the reference point. "
                "(type: numpy.ndarray)"
            )
        }

    @property
    def I_epsilon(self):
        return self.__I_epsilon

    @property
    def I_alpha(self):
        return self.__I_alpha

    @property
    def mu(self):
        return self.__mu

    @property
    def w(self):
        return self.__w

    @property
    def q(self):
        return self.preference["reference point"]

    @property
    def epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def s_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def delta_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)


class GUESS_GLIDE(GLIDEBase):
    """
    Implements the GUESS method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):
        super().__init__(utopian=utopian, nadir=nadir, rho=rho)
        self.has_additional_constraints = False
        self.__I_alpha = np.full_like(
            utopian, dtype=np.bool_, fill_value=True
        ).flatten()
        self.__I_epsilon = np.full_like(
            utopian, dtype=np.bool_, fill_value=False
        ).flatten()
        self.__w = 0
        self.required_keys = {
            "reference point": (
                "Used to calculate the direction of improvement: "
                "a line going from the nadir point to the reference point. "
                "(type: numpy.ndarray)"
            )
        }

    @property
    def I_epsilon(self):
        return self.__I_epsilon

    @property
    def I_alpha(self):
        return self.__I_alpha

    @property
    def mu(self):
        return 1 / (self.nadir - self.preference["reference point"])

    @property
    def w(self):
        return self.__w

    @property
    def q(self):
        return self.preference["reference point"]

    @property
    def epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def s_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def delta_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)


class AUG_GUESS_GLIDE(GUESS_GLIDE):
    """
    Implements the Augmented GUESS method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):
        super().__init__(utopian=utopian, nadir=nadir, rho=rho)
        self.__w = 1


class NIMBUS_GLIDE(GLIDEBase):
    """
    Implements the NIMBUS method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):
        super().__init__(utopian=utopian, nadir=nadir, rho=rho)

        self.__mu = self.__w = 1 / (self.nadir - self.utopian)

        self.has_additional_constraints = True
        self.required_keys = {
            "current solution": (
                "A solution preferred by the DM currently. " "(type: numpy.ndarray)"
            ),
            "classifications": (
                "A list of same length as the number of objectives. Elements can only "
                "include some or all of ['<', '<=', '=', '>=', '0']. These classify "
                "the different objectives as defined in the NIMBUS or GLIDE-II paper. "
                "(type: list)"
            ),
            "levels": (
                "A vector containing desirable levels of objectives or constraining bounds "
                "depending on the classification. Same length as the number of objectives. "
                "(type: numpy.ndarray)"
            ),
        }

    @property
    def improve_unconstrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "<")[0]
        indices[relevant] = True
        return indices

    @property
    def improve_constrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "<=")[0]
        indices[relevant] = True
        return indices

    @property
    def satisfactory(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "=")[0]
        indices[relevant] = True
        return indices

    @property
    def relax_constrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == ">=")[0]
        indices[relevant] = True
        return indices

    @property
    def relax_unconstrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "0")[0]
        indices[relevant] = True
        return indices

    @property
    def I_alpha(self):
        return self.improve_unconstrained + self.improve_constrained

    @property
    def I_epsilon(self):
        return (
            self.improve_unconstrained
            + self.improve_constrained
            + self.satisfactory
            + self.relax_constrained
        )

    @property
    def w(self):
        # This was in the paper
        return self.__w
        # This is what I think it should be. There may be division by zero errors here.
        """return (self.objective_vector / (self.objective_vector - self.q)) / (
            self.nadir - self.utopian
        )"""

    @property
    def mu(self):
        return self.__mu

    @property
    def q(self):
        q = np.full_like(self.utopian, fill_value=0, dtype=float)
        q[self.improve_unconstrained] = self.utopian[self.improve_unconstrained]
        q[self.improve_constrained] = self.preference["levels"][
            self.improve_constrained
        ]
        return q

    @property
    def epsilon(self):
        e = np.full_like(self.utopian, fill_value=np.nan, dtype=float)

        case1 = (
            self.improve_constrained + self.improve_unconstrained + self.satisfactory
        )
        case2 = self.relax_constrained

        e[case1] = self.preference["current solution"][case1]
        e[case2] = self.preference["levels"][case2]
        return e

    @property
    def s_epsilon(self):
        return 0

    @property
    def delta_epsilon(self):
        return np.zeros_like(self.utopian)


class STEP_GLIDE(GLIDEBase):
    """
    Implements the STEP method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(
        self, utopian: np.ndarray = None, nadir: np.ndarray = None, rho: float = 1e-6
    ):
        super().__init__(utopian=utopian, nadir=nadir, rho=rho)
        self.__mu = (self.nadir - self.utopian) / np.max(
            np.abs(np.vstack((utopian, nadir))), axis=0
        )
        self.__w = 0

        self.I_epsilon = np.full_like(self.utopian, dtype=np.bool_, fill_value=True)

        self.has_additional_constraints = True
        self.required_keys = {
            "current solution": (
                "A solution preferred by the DM currently. " "(type: numpy.ndarray)"
            ),
            "classifications": (
                "A list of same length as the number of objectives. Elements can only "
                "include some or all of [<=', '=', '>=']. These classify "
                "the different objectives as defined in the GLIDE-II paper. "
                "(type: list)"
            ),
            "levels": (
                "A vector containing desirable levels of objectives or constraining bounds "
                "depending on the classification. Same length as the number of objectives. "
                "(type: numpy.ndarray)"
            ),
        }

    @property
    def improve_constrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "<=")[0]
        indices[relevant] = True
        return indices

    @property
    def satisfactory(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == "=")[0]
        indices[relevant] = True
        return indices

    @property
    def relax_constrained(self):
        indices = np.full_like(self.utopian, dtype=np.bool_, fill_value=False)
        relevant = np.where(np.array(self.preference["classifications"]) == ">=")[0]
        indices[relevant] = True
        return indices

    @property
    def I_alpha(self):
        return self.improve_constrained

    @property
    def w(self):
        # This was in the paper
        return self.__w

    @property
    def mu(self):
        return self.__mu

    @property
    def q(self):
        q = np.full_like(self.utopian, fill_value=0, dtype=float)
        q[self.improve_constrained] = self.utopian[self.improve_constrained]
        return q

    @property
    def epsilon(self):
        e = np.full_like(self.utopian, fill_value=np.nan, dtype=float)

        case1 = self.improve_constrained + self.satisfactory
        case2 = self.relax_constrained

        e[case1] = self.preference["current solution"][case1]
        e[case2] = self.preference["levels"][case2]
        return e

    @property
    def s_epsilon(self):
        return 0

    @property
    def delta_epsilon(self):
        return np.zeros_like(self.utopian)


class STOM_GLIDE(GLIDEBase):
    """
    Implements the STOM method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(self, utopian: np.ndarray = None, rho: float = 1e-6):
        super().__init__(utopian=utopian, nadir=None, rho=rho)
        self.has_additional_constraints = False
        self.__I_alpha = np.full_like(
            utopian, dtype=np.bool_, fill_value=True
        ).flatten()
        self.__I_epsilon = np.full_like(
            utopian, dtype=np.bool_, fill_value=False
        ).flatten()
        self.__w = 0
        self.required_keys = {
            "reference point": (
                "Used to calculate the direction of improvement: "
                "a line going from the reference point to the utopian point. "
                "(type: numpy.ndarray)"
            )
        }

    @property
    def I_epsilon(self):
        return self.__I_epsilon

    @property
    def I_alpha(self):
        return self.__I_alpha

    @property
    def mu(self):
        return 1 / (self.preference["reference point"] - self.utopian)

    @property
    def w(self):
        return self.__w

    @property
    def q(self):
        return self.utopian

    @property
    def epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def s_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def delta_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)


class AUG_STOM_GLIDE(STOM_GLIDE):
    """
    Implements the Augmented STOM method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(self, utopian: np.ndarray = None, rho: float = 1e-6):
        super().__init__(utopian=utopian, rho=rho)
        self.has_additional_constraints = False
        self.__w = 1


class Tchebycheff_GLIDE(GLIDEBase):
    """
    Implements the Tchebycheff method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(self, utopian: np.ndarray = None, rho: float = 1e-6):
        super().__init__(utopian=utopian, nadir=None, rho=rho)
        self.has_additional_constraints = False
        self.__I_alpha = np.full_like(
            utopian, dtype=np.bool_, fill_value=True
        ).flatten()
        self.__I_epsilon = np.full_like(
            utopian, dtype=np.bool_, fill_value=False
        ).flatten()
        self.__w = 1
        self.required_keys = {
            "mu": (
                "Vector defining the direction of improvement of the scalarizer. "
                "(type: numpy.ndarray)"
            )
        }

    @property
    def I_epsilon(self):
        return self.__I_epsilon

    @property
    def I_alpha(self):
        return self.__I_alpha

    @property
    def mu(self):
        return self.preference["mu"]

    @property
    def w(self):
        return self.__w

    @property
    def q(self):
        return self.utopian

    @property
    def epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def s_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def delta_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)


class PROJECT_GLIDE(GLIDEBase):
    """
    Implements the PROJECT method of preference elicitation and scalarization
    using the non-differentiable variant of GLIDE-II as proposed in:
    Ruiz, Francisco, Mariano Luque, and Kaisa Miettinen.
    "Improving the computational efficiency in a global formulation (GLIDE)
    for interactive multiobjective optimization."
    Annals of Operations Research 197.1 (2012): 47-70.

    Args:
        utopian (np.ndarray, optional): The utopian point. Defaults to None.
        nadir (np.ndarray, optional): The nadir point. Defaults to None.
        rho (float, optional): The augmentation term for the scalarization function.
            Defaults to 1e-6.
    """

    def __init__(self, current_objective_vector: np.ndarray, rho: float = 1e-6):
        super().__init__(utopian=None, nadir=None, rho=rho)
        self.current_objective_vector = current_objective_vector
        self.has_additional_constraints = False
        self.__I_alpha = np.full_like(
            current_objective_vector, dtype=np.bool_, fill_value=True
        ).flatten()
        self.__I_epsilon = np.full_like(
            current_objective_vector, dtype=np.bool_, fill_value=False
        ).flatten()
        self.__w = 0

    @property
    def I_epsilon(self):
        return self.__I_epsilon

    @property
    def I_alpha(self):
        return self.__I_alpha

    @property
    def mu(self):
        return 1 / np.abs(
            self.preference["reference point"] - self.current_objective_vector
        )

    @property
    def w(self):
        return self.__w

    @property
    def q(self):
        return self.utopian

    @property
    def epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def s_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)

    @property
    def delta_epsilon(self):
        msg = "This part of the code should not be reached. Contact maintaner."
        raise GLIDEError(msg)



"""
This module implements various small tools that may be useful during optimization using
either the MCDM or the EA methods.
"""

__all__ = [
    "dominates",
    "non_dominated",
    "fast_non_dominated_sort",
    "fast_non_dominated_sort_indices",
    "classification_to_reference_point",
]

from desdeo_tools.utilities.fast_non_dominated_sorting import (
    dominates,
    non_dominated,
    fast_non_dominated_sort,
    fast_non_dominated_sort_indices,
)

from desdeo_tools.utilities.preference_converters import (
    classification_to_reference_point,
)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['desdeo_tools',
 'desdeo_tools.interaction',
 'desdeo_tools.scalarization',
 'desdeo_tools.solver',
 'desdeo_tools.utilities']

package_data = \
{'': ['*']}

install_requires = \
['numba>=0.53.1,<0.54.0',
 'numpy>=1.17,<2.0',
 'pandas>=1.0,<2.0',
 'scipy>=1.2,<2.0']

setup_kwargs = {
    'name': 'desdeo-tools',
    'version': '1.4.1',
    'description': 'Generic tools and design language used in the DESDEO framework',
    'long_description': None,
    'author': 'Bhupinder Saini',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
